
import java.util.*;
import java.io.*;
import parser.*;
import parser.Syntax.*;
import parser.Parser.*;

public class JavaMerger {
    public static void main(String[] args) {
        try {
            if (args.length == 0) {
                System.out.println("usage: JavaMerger files...");
                return;
            }
            
            mergeFiles(asList(args), null, new OutputStreamWriter(System.out));
        } catch (Exception x) {
            x.printStackTrace();
        }
    }
    
    private static List asList(Object[] array) {
        return Arrays.asList(array);
    }
    private static List asList(List list) {
        return list;
    }
    
    private static class EditeeCompilationUnit extends Default.Node implements CompilationUnit {
        public EditeeCompilationUnit() {
            super(new LinkedList());
        }
    }
    public static void mergeFiles(Collection sources, String encoding, Writer writer) throws IOException, ParseException {
        Parser parser = new Parser(true, true);
        //CompilationUnit unit = null;
        CompilationUnit unit = new EditeeCompilationUnit();
        Iterator it2 = sources.iterator();
        while (it2.hasNext()) {
            Object source = (Object) it2.next();
            File f = (source instanceof File) ? (File) source : new File((String) source);
            CompilationUnit unit2 = parser.parseCompilationUnit(f.getPath(), f, encoding, 8);
            if (unit == null) {
                unit = unit2;
            } else {
                merge(unit, unit2);
            }
        }
        new JavaMerger().output(unit, writer);
    }
    
    private static void merge(Node node1, Node node2) {
        // merge node2 into node1
        if (node1 instanceof CompilationUnit) {
            // merge PackageDeclaration
            merge(new ClassChecker(PackageDeclaration.class), new ClassChecker(Node.class), new IndexFinder() {
                public int getIndex(Node node) {
                    int result = 0;
                    Iterator it3 = node.getChildNodes().iterator();
                    for (int i = 0; it3.hasNext(); i++) {
                        Node child = (Node) it3.next();
                        if (child instanceof PackageDeclaration)
                            result = i + 1;
                    }
                    return result;
                }
            }, node1, node2);
            
            // merge import declarations
            IndexFinder finder = new IndexFinder() {
                public int getIndex(Node node) {
                    int result = 0;
                    Iterator it3 = node.getChildNodes().iterator();
                    for (int i = 0; it3.hasNext(); i++) {
                        Node child = (Node) it3.next();
                        if (child instanceof PackageDeclaration || child instanceof ImportDeclaration)
                            result = i + 1;
                    }
                    return result;
                }
            };
            merge(new ClassChecker(SingleTypeImportDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(TypeImportOnDemandDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(SingleStaticImportDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(StaticImportOnDemandDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            
            // merge type declaration
            merge(new ClassChecker(TypeDeclaration.class), new ClassChecker(Identifier.class), null, node1, node2);
            
            // IntegratedMethodDeclaration
            //unintegrateMethod(node1, node2);
        } else if (node1 instanceof TypeDeclaration) {
            // merge modifiers
            merge(new ClassChecker(Token.class), null, null, find(node1, Modifiers.class), find(node2, Modifiers.class));
            
            // merge inheritance
            merge(find(node1, Extendees.class), find(node2, Extendees.class));
            merge(find(node1, Implementees.class), find(node2, Implementees.class));
            
            // merge nested type declarations
            merge(new ClassChecker(TypeDeclaration.class), new ClassChecker(Identifier.class), null, find(node1, Body.class), find(node2, Body.class));
            
            // merge the other members
            merge(new Checker() {
                public boolean hit(Node node) {
                    return !(node instanceof TypeDeclaration);
                }
            }, new Checker() {
                public boolean hit(Node node) {
                    return false;
                }
            }, null, find(node1, Body.class), find(node2, Body.class));
            
            // IntegratedMethodDeclaration
            //unintegrateMethod(find(node1, Body.class), find(node2, Body.class));
        } else if (node1 instanceof Extendees) {
            Extendees n1 = (Extendees) node1;
            Extendees n2 = (Extendees) node2;
            Iterator it1 = asList(n2.extendees()).iterator();
            while (it1.hasNext()) {
                ClassOrInterfaceType extendee = (ClassOrInterfaceType) it1.next();
                if (n1.getChildNodes().isEmpty()) {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_EXTENDS, "extends"));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                } else {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_COMMA, ","));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                }
                n1.getChildNodes().add(extendee);
            }
        } else if (node1 instanceof Implementees) {
            Implementees n1 = (Implementees) node1;
            Implementees n2 = (Implementees) node2;
            Iterator it1 = asList(n2.implementees()).iterator();
            while (it1.hasNext()) {
                ClassOrInterfaceType implementee = (ClassOrInterfaceType) it1.next();
                if (n1.getChildNodes().isEmpty()) {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_IMPLEMENTS, "implements"));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                } else {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_COMMA, ","));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                }
                n1.getChildNodes().add(implementee);
            }
        }
    }
    private static Node find(Node node, Class clazz) {
        Iterator it9 = node.getChildNodes().iterator();
        while (it9.hasNext()) {
            Node child = (Node) it9.next();
            if (clazz.isInstance(child))
                return child;
        }
        return null;
    }
    
    private static interface Checker {
        public boolean hit(Node node);
    }
    private static class ClassChecker implements Checker {
        private Class clazz;
        public ClassChecker(Class clazz) {
            this.clazz = clazz;
        }
        public boolean hit(Node node) {
            return clazz.isInstance(node);
        }
    }
    private static interface IndexFinder {
        public int getIndex(Node node);
    }
    
    private static void merge(Checker targetChecker, Checker identifierChecker, IndexFinder finder, Node node1, Node node2) {
        Map idToNode = new LinkedHashMap();
        {
            Iterator it3 = node1.getChildNodes().iterator();
            while (it3.hasNext()) {
                Node child = (Node) it3.next();
                if (targetChecker.hit(child)) {
                    StringBuffer id = new StringBuffer();
                    if (identifierChecker == null) {
                        id.append(toEssentialString(child));
                    } else {
                        Iterator it4 = child.getChildNodes().iterator();
                        while (it4.hasNext()) {
                            Node grandchild = (Node) it4.next();
                            if (identifierChecker.hit(grandchild))
                                id.append(toEssentialString(grandchild));
                        }
                    }
                    idToNode.put(id.toString(), child);
                }
            }
        }
        
        {
            Token lastJavadoc = null;
            Iterator it3 = node2.getChildNodes().iterator();
            while (it3.hasNext()) {
                Node child = (Node) it3.next();
                if (child instanceof Token && ((Token) child).isWhite()) {
                    Token token = (Token) child;
                    if (token.getSymbolID() == Syntax.TOKEN_JAVADOC_COMMENT) {
                        lastJavadoc = token;
                    }
                } else if (!targetChecker.hit(child)) {
                    lastJavadoc = null;
                } else {
                    StringBuffer id = new StringBuffer();
                    if (identifierChecker == null) {
                        id.append(toEssentialString(child));
                    } else {
                        Iterator it4 = child.getChildNodes().iterator();
                        while (it4.hasNext()) {
                            Node grandchild = (Node) it4.next();
                            if (identifierChecker.hit(grandchild))
                                id.append(toEssentialString(grandchild));
                        }
                    }
                    Node earlierVisitor = (id.length() == 0) ? null : (Node) idToNode.get(id.toString());
                    if (earlierVisitor == null) {
                        int index = node1.getChildNodes().size();
                        if (finder != null) {
                            index = finder.getIndex(node1);
                        }
                        Object[] list;
                        if (lastJavadoc != null)
                            list = new Object[] { lastJavadoc, child, new Default.Token(Syntax.TOKEN_WHITE_SPACES, "\n") };
                        else
                            list = new Object[] { new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "), child, new Default.Token(Syntax.TOKEN_WHITE_SPACES, "\n") };
                        node1.getChildNodes().addAll(index, asList(list));
                        if (id.length() != 0)
                            idToNode.put(id.toString(), child);
                    } else {
                        merge(earlierVisitor, child);
                    }
                }
            }
        }
    }
    
    private static String toEssentialString(Node node) {
        StringBuffer buffer = new StringBuffer();
        toEssentialString(node, buffer);
        return buffer.toString();
    }
    private static void toEssentialString(Node node, StringBuffer buffer) {
        if (node instanceof Token) {
            Token token = (Token) node;
            if (!token.isWhite())
                buffer.append(((Token) node).getOriginalImage());
            return;
        }
        
        Iterator it0 = node.getChildNodes().iterator();
        while (it0.hasNext()) {
            Node child = (Node) it0.next();
            toEssentialString(child, buffer);
        }
    }
    
    private boolean outputPosition = true;
    private void output(Node node, Writer writer) throws IOException {
        writer.write("/*\n");
        outputPosition = true;
        outputNode(node, writer);
        if (outputPosition) {
            writer.write("*/");
        }
        writer.flush();
    }
    private void outputNode(Node node, Writer writer) throws IOException {
        if (node instanceof Token) {
            Token token = (Token) node;
            if (outputPosition) {
                try {
                    writer.write(token.getSourceName());
                    writer.write(':');
                    writer.write(Integer.toString(token.getLine()));
                    writer.write(':');
                } catch (UnsupportedOperationException x) {
                }
                writer.write("*/");
                outputPosition = false;
            }
            if (token.getSymbolID() == Parser.TOKEN_LINE_TERMINATOR) {
                writer.write("/*");
                outputPosition = true;
            }
            writer.write(token.getOriginalImage());
            return;
        }
        
        /*if (node instanceof EnhancedForStatement) {
            outputNode((EnhancedForStatement) node, writer);
            return;
        }*/
        
        Iterator it0 = node.getChildNodes().iterator();
        while (it0.hasNext()) {
            Node child = (Node) it0.next();
            outputNode(child, writer);
        }
    }
    /*private void outputNode(EnhancedForStatement node, Writer writer) throws IOException {
        outputNode(node.forKeyword(), writer);
        outputNode(node.leftParen(), writer);
        {
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "java.util.Iterator $$i$$ = ("), writer);
            outputNode(node.expression(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ").iterator(); $$i$$.hasNext();"), writer);
        }
        outputNode(node.rightParen(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "{"), writer);
        
        outputNode(node.type(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " "), writer);
        outputNode(node.identifier(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " = ("), writer);
        outputNode(node.type(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ") $$i$$.next();"), writer);
        
        outputNode(node.statement(), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "}"), writer);
    }*/
}

